# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Pacman;
use strict;
use warnings;

use Xyne::Arch::PacmanConf;
use Xyne::Common::Shell;
use Xyne::Common::Functions qw/dejuxtapose_args format_bytes hash_map wrap_line/;

our $VERSION = '0.33';

my $arg_maps = {
  OPERATIONS => {
    '--query'   => 'QUERY',
    '--remove'  => 'REMOVE',
    '--sync'    => 'SYNC',
    '--upgrade' => 'UPGRADE',
    '--version' => 'VERSION',
    '--help'    => 'HELP'
  },
  PARAMETRIC => {
    '--dbpath'      => 'single',
    '--cachedir'    => 'multiple',
    '--config'      => 'single',
    '--ignore'      => 'multiple',
    '--ignoregroup' => 'multiple',
    '--logfile'     => 'single',
    '--owns'        => 'single',
    '--root'        => 'single',
    '--search'      => 'single'
  },
  CONF => {
    '--dbpath'      => 'DBPath',
    '--cachedir'    => 'CacheDir',
    '--ignore'      => 'IgnorePkg',
    '--ignoregroup' => 'IgnoreGroup',
    '--logfile'     => 'LogFile',
    '--root'        => 'RootDir',
    '--config'      => 'Config'
  },
  EXPANSIONS => {
    COMMON => {
      '-Q' => '--query',
      '-R' => '--remove',
      '-S' => '--sync',
      '-U' => '--upgrade',
      '-V' => '--version',
      '-h' => '--help',
      '-b' => '--dbpath',
      '-d' => '--nodeps',
      '-f' => '--force',
      '-r' => '--root',
      '-v' => '--verbose'
    },
    QUERY => {
      '-c' => '--changelog',
      # this is needed to handle the overlapping abbreviation "-d"
      '-d' => '--deps', '--nodeps' => '--deps',
      '-e' => '--explicit',
      '-g' => '--groups',
      '-i' => '--info',
      '-l' => '--list',
      '-m' => '--foreign',
      '-o' => '--owns',
      '-p' => '--file',
      '-q' => '--quiet',
      '-s' => '--search',
      '-t' => '--unrequired',
      '-u' => '--upgrades'
    },
    REMOVE => {
      '-c' => '--cascade',
      '-k' => '--keep',
      '-n' => '--nosave',
      '-s' => '--recursive',
      '-u' => '--unneeded'
    },
    SYNC => {
      '-c' => '--clean',
      '-g' => '--groups',
      '-i' => '--info',
      '-l' => '--list',
      '-p' => '--print-uris',
      '-q' => '--quiet',
      '-s' => '--search',
      '-u' => '--sysupgrade',
      '-w' => '--downloadonly',
      '-y' => '--refresh'
    }
  }
};

my $REFRESH_FILTER = {
  'add'  => [qw/--noprogressbar/],
  'list' => [qw/--sync --refresh --quiet --dbpath --cachedir --logfile --root --config/],
  'type' => 'inclusive'
};

my $NOREFRESH_FILTER = {
  'list' => [qw/--refresh/],
  'type' => 'exclusive'
};

my @NODOWNLOAD = qw/--clean --groups --info --list --print-uris --search/;
my @NEEDSROOT = qw/--clean --print-uris --refresh --remove --sysupgrade --upgrade/;


sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);

  $self->{BINARY} = $args->{BINARY} ? $args->{BINARY} : '/usr/bin/pacman';
  if (defined($args->{PacmanConf}))
  {
    $self->{PacmanConf} = $args->{PacmanConf};
  }
  else
  {
    $self->{PacmanConf} = Xyne::Arch::PacmanConf->new( $args->{PacmanConf_PATH} );
  }
  $self->{OPERATION} = $args->{OPERATION} ? $args->{OPERATION} : 'NONE';
  $self->{PKGS} = $args->{PKGS} ? $args->{PKGS} : [];
  $self->{ARGS}->{ARRAY} = [];
  $self->{DB} = {};
  $self->{Shell} = $args->{Shell} ? $args->{Shell} : Xyne::Common::Shell->new();
  $self->{WRAP_WIDTH} = 80;

  my ($sudo_binary) = $self->{Shell}->execute('which sudo 2>/dev/null');
  if (defined($sudo_binary))
  {
    chomp $sudo_binary;
  }
  else
  {
    $sudo_binary = '';
  }
  $self->{SUDO_BINARY} = $sudo_binary;

  return $self;
}








sub add_to_download
{
  my ($self,$pkg,$download) = @_;
  if (not defined $download->{$pkg} and $pkg)
  {
    #$download->{$self->strip_ver_spec($pkg)} ++;
    $download->{$pkg} ++;
    return $download if $self->{ARGS}->{'--nodeps'};
    foreach my $dep ($self->get_sync_db_deps($pkg))
    {
      next if $self->is_in_local_db($dep);
      if (my $target = $self->is_in_sync_db($dep))
      {
        $self->add_to_download($target,$download);
      }
      else
      {
        my $provider = $self->get_sync_db_provider($dep);
        if (defined($provider))
        {
          $download = $self->add_to_download($provider,$download);
        }
      }
    }
  }
  return $download;
}

sub add_pkg
{
  my ($self,$pkg) = @_;
  $pkg = $self->maybe_prepend_repo($pkg);
  push @{$self->{PKGS}}, $pkg;
}

sub append_csize_tags
{
  my ($self,@pkgs) = @_;
  return map { $_ . ' [' . &format_bytes($self->get_sync_db_csize($_),'MiB') . ']' } @pkgs;
}

sub append_isize_tags
{
  my ($self,@pkgs) = @_;
  return map { $_ . ' [' . &format_bytes($self->get_sync_db_isize($_),'MiB') . ']' } @pkgs;
}

sub check_cachedirs_for_existing_files
{
  my ($self,$download) = @_;
  my @pkgs = keys %{ $download };
  my @files;
  my %file_to_pkg;
  foreach my $cache ($self->{PacmanConf}->get_option('CacheDir'))
  {
    foreach my $pkg (@pkgs)
    {
      my $file = $cache.'/'.$self->get_sync_db_filename($pkg);
      if (-e $file)
      {
        push @files,$file;
        $file_to_pkg{$file} = $pkg;
      }
    }
  }
  return $download if not @files;
  foreach my $line ($self->get_disk_md5sums(@files))
  {
    if ($line =~ m/^([0-9A-Fa-f]{32})\s+(\S.+)$/)
    {
      my ($md5sum,$file) = ($1,$2);
      my $pkg = $file_to_pkg{$file};
      if ($md5sum eq $self->get_sync_db_md5sum($pkg))
      {
        delete $download->{$pkg};
      }
      else
      {
        # If the aria2c file is there and this is being run by powerpill, keep the file to resume downloading.
        # I'll add some generic option later to remove the powerpill/aria2 specificity.
        next if -e $file.'.aria2' and defined($self->{Powerpill});
        unlink $file if $self->confirm(":: File $file is corrupted. Do you want to delete it?",'y');
      }
    }
  }
  return $download;
}

sub clear_arg
{
  my ($self,$name) = @_;
  $name = $self->expand($name);
  if (defined($self->{ARGS}->{$name}))
  {
    delete ($self->{ARGS}->{$name});
  }
  my @args = @{ $self->{ARGS}->{ARRAY} };
  my @remaining_args;
  while (@args)
  {
    my $arg = shift @args;
    my $remove = $arg eq $name;
    push @remaining_args,$arg if not $remove;

    if (defined($arg_maps->{PARAMETRIC}->{$arg}))
    {
      my $parameter = @args ? shift @args : '';
      push @remaining_args,$parameter if not $remove;
    }
  }
  $self->{ARGS}->{ARRAY} = [@remaining_args];
}

sub confirm
{
  my ($self,$question,$default) = @_;
  return lc($default) eq 'y' if ($self->{ARGS}->{'--noconfirm'});
  return &Xyne::Common::Functions::confirm($question,$default);
}

# stick this here to handle message suppression etc
sub display
{
  my ($self,$text) = @_;
  print $text;
}

sub error
{
  my ($self,@output) = @_;
  print STDERR @output;
  exit 1;
}

sub execute
{
  my ($self,$filter,$exit) = @_;
  my @args = $self->get_args($filter);
  my $cmd = $self->{BINARY}.' '.join(' ',@args);
  if ($self->requires_root(@args) and $ENV{'USER'} ne 'root')
  {
    if ($self->{SUDO_BINARY} eq '')
    {
      #$self->error("error: this operation requires root privileges\n","Either log in as root and rerun the application\n","or install and configure \"sudo\".\n")
      $self->warning("warning: this operation requires root privileges\n", "You are currently running this application as $ENV{'USER'}.\n","If the operation does not succeed, either log in as root\nor install and configure sudo and then try again.\n\n")
    }
    else
    {
      $cmd = 'sudo '.$cmd;
    }
  }
  if ($exit)
  {
    exec($cmd);
    exit;
  }
  my @output = $self->{Shell}->execute($cmd);
  if ($self->get_exit_status)
  {
    $self->error(@output);
  }
  else
  {
    return @output;
  }
}

sub expand
{
  my ($self,$arg) = @_;
  return defined($self->{EXPANSIONS}->{$arg}) ? $self->{EXPANSIONS}->{$arg} : $arg;
}

sub get_arg
{
  my ($self,$arg) = @_;
  $arg = $self->expand($arg);
  if (defined($self->{ARGS}->{$arg}))
  {
    return $self->{ARGS}->{$arg};
  }
  elsif (defined($arg_maps->{PARAMETRIC}->{$arg}))
  {
    return '';
  }
  else
  {
    return 0;
  }
}

sub get_args_pkgs
{
  my ($self) = @_;
  return () if not defined($self->{PKGS});
  my $pkgs = $self->{PKGS};
  return (ref($pkgs) eq 'ARRAY') ? @$pkgs : ($pkgs);
}

sub get_args
{
  my ($self,$filter) = @_;
  my @args = @{ $self->{ARGS}->{ARRAY} };
  my %filter;
  $filter{$_} = 1 foreach @args;
  my $threshold = 0;
  my @filtered_args = ();
  if (defined($filter->{'list'}) and defined($filter->{'type'}))
  {
    if ($filter->{'type'} eq 'inclusive')
    {
      $threshold = 1;
      foreach my $item (@{ $filter->{'list'} })
      {
        $filter{$self->expand($item)}++;
      }
    }
    elsif ($filter->{'type'} eq 'exclusive')
    {
      foreach my $item (@{ $filter->{'list'} })
      {
        $filter{$self->expand($item)}=0;
      }
    }
  }
  while (@args)
  {
    my $arg = shift @args;
    my $parameter;
    if (defined($arg_maps->{PARAMETRIC}->{$arg}))
    {
      $parameter = @args ? shift @args : '';
    }
    if ($filter{$arg} > $threshold)
    {
      push @filtered_args, $arg;
      push @filtered_args, $parameter if defined($parameter);
    }
  }
  if (defined($filter->{'add'}))
  {
    my @additional_args = @{ $filter->{'add'} };
    push @filtered_args, @additional_args;
  }
  return @filtered_args;
}

sub get_arg_str
{
  my ($self,$filter) = @_;
  # join arguments by spaces, wrapping arguments with spaces in them with quoation marks
  return join(' ', map {(index($_,' ') > -1) ? "\"$_\"" : $_} $self->get_args($filter));
}

sub get_binary
{
  my ($self) = @_;
  return $self->{BINARY};
}

sub get_sync_db_csize
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_scalar($pkg,'CSIZE');
}

sub get_sync_db_deps
{
  my ($self,$pkg) = @_;
  return map {$self->maybe_prepend_repo($_)} $self->get_sync_db_field_as_array($pkg,'DEPENDS');
}

sub get_disk_md5sums
{
  my ($self,@files) = @_;
  @files = grep {-f} @files;
  return '' if not @files;
  return $self->{Shell}->execute('md5sum '.join(' ',@files));
}

sub get_download
{
  my ($self) = @_;
  $self->load_local_db if not defined($self->{DB}->{LOCAL});
  $self->load_sync_db if not defined($self->{DB}->{SYNC});

  # "_from_binary" can be removed
  my $download = $self->get_sysupgrade_download_from_binary if $self->{ARGS}->{'--sysupgrade'};
  foreach my $pkg ($self->get_args_pkgs)
  {
    if (not $self->is_in_sync_db($pkg))
    {
      my $provider = $self->get_sync_db_provider($pkg);
      if (defined($provider))
      {
        $download = $self->add_to_download($provider,$download);
      }
      else
      {
        $self->display("$pkg package not found, searching for group...\n");
        my @group = $self->get_sync_db_group_pkgs($pkg);
        if (@group)
        {
          my $question = ":: group $pkg (including ignored packages):\n";
          my @names = @group;
          if ($self->{PacmanConf}->get_option('ShowSize'))
          {
            @names = $self->append_csize_tags(@names);
          }
          @names = map {$self->get_name($_)} @names;
          $question .= &wrap_line($self->{WRAP_WIDTH},'    ',@names);
          $question .= ":: Install whole content?";
          if ($self->confirm($question,'y'))
          {
            foreach my $member (@group)
            {
              $download = $self->add_to_download($member,$download);
            }
          }
          else
          {
            foreach my $member (@group)
            {
              next if defined($download->{$member});
              my $question = ":: Install ".$self->get_name($member)." from group $pkg?";
              $download = $self->add_to_download($member,$download) if $self->confirm($question,'y');
            }
          }
        }
        else
        {
          $self->error("error: '$pkg': not found in sync db\n");
        }
      }
    }
    else
    {
      $download = $self->add_to_download($pkg,$download);
    }
  }
  my @sync_first = $self->get_sync_first;
  my $sync_first_download;
  my @sync_first_queue;
  if (not defined($self->{ARGS}->{'--downloadonly'}))
  {
    foreach my $name (@sync_first)
    {
      my $pkg = $self->maybe_prepend_repo($name);
      next if not defined($download->{$pkg}) or not $self->is_in_sync_db($pkg);
      if (not $self->is_in_local_db($name)
        or &compare_versions(
          $self->get_sync_db_ver($pkg),
          $self->get_local_db_ver($name)
        ) == 1)
      {
        $sync_first_download->{$pkg} = $download->{$pkg};
        push @sync_first_queue, $pkg;
      }
    }
  }
  if (@sync_first_queue and scalar(keys(%{$download})) > scalar(@sync_first_queue))
  {
    my $question = ":: The following packages should be upgraded first:\n";
    $question .= '    '.join(' ', @sync_first_queue)."\n";
    $question .= "::Do you want to cancel the current operation\n";
    $question .= "::and upgrade these packages now?";
    $download = $sync_first_download if $self->confirm($question,'y');
  }
  return $self->check_cachedirs_for_existing_files($download);
}

sub get_exit_status
{
  my ($self) = @_;
  return $self->{Shell}->{EXIT_STATUS};
}

sub get_local_db_field
{
  my ($self,$pkg,$field) = @_;
  $self->load_local_db_file($pkg,'desc');
  $self->load_local_db_file($pkg,'depends');
  my $pkg_data = $self->{DB}->{LOCAL}->{$pkg};
  return  $pkg_data->{$field};
}

sub get_local_db_field_as_array
{
  my ($self,$pkg,$field) = @_;
  my $value = $self->get_local_db_field($pkg,$field);
  return grep {$_} (ref($value) eq 'ARRAY') ? @$value : ($value);
}

sub get_local_db_field_as_scalar
{
  my ($self,$pkg,$field) = @_;
  my $value = $self->get_local_db_field($pkg,$field);
  return (ref($value) eq 'ARRAY') ? $value->[0] : $value;
}

sub get_ignored_groups
{
  my ($self) = @_;
  return $self->{PacmanConf}->get_option('IgnoreGroup');
}

sub get_ignored_pkgs
{
  my ($self) = @_;
  return $self->{PacmanConf}->get_option('IgnorePkg');
}

sub get_local_db_data
{
  my ($self,$pkg) = @_;
  return $self->{DB}->{LOCAL}->{$pkg};
}

sub get_local_db_provider
{
  my ($self,$name) = @_;
  $self->load_local_db if not defined($self->{DB}->{LOCAL});
  foreach my $pkg ($self->get_local_db_pkgs)
  {
    return $pkg if ($self->is_in_local_db_provides($pkg,$name));
  }
  return undef;
}

sub get_local_db_provides
{
  my ($self,$pkg) = @_;
  return $self->get_local_db_field_as_array($pkg,'PROVIDES');
}

sub get_local_db_pkgs
{
  my ($self) = @_;
  return keys %{ $self->{DB}->{LOCAL} };
}

sub get_local_db_ver
{
  my ($self,$pkg) = @_;
  return $self->get_local_db_data($pkg)->{VERSION};
}

sub get_minimum_version
{
  my ($self,$pkg) = @_;
  if ($pkg =~ s/(?:>|=|>=)(.+)$//)
  {
    return $1;
  }
  return 0;
}

sub get_name
{
  my ($self,$pkg) = @_;
  if ($pkg =~ m/^[^\/]+\/([^\/]+)$/)
  {
    $pkg = $1;
  }
  return $pkg;
}

sub get_name_and_min_ver
{
   my ($self,$pkg) = @_;
   if ($pkg =~ m/^(.+?)(?:>|=|>=)(.+)$/)
   {
     return $1,$2;
   }
   else
   {
      return $pkg,undef;
   }
}

sub get_sync_db_data
{
  my ($self,$pkg) = @_;
  return $self->{DB}->{SYNC}->{$pkg};
}

sub get_sync_db_field
{
  my ($self,$pkg,$field) = @_;
  $self->load_sync_db_file($pkg,'desc');
  $self->load_sync_db_file($pkg,'depends');
  my $pkg_data = $self->{DB}->{SYNC}->{$pkg};
  return  $pkg_data->{$field};
}

sub get_sync_db_field_as_array
{
  my ($self,$pkg,$field) = @_;
  my $value = $self->get_sync_db_field($pkg,$field);
  return grep {$_} (ref($value) eq 'ARRAY') ? @$value : ($value);
}

sub get_sync_db_field_as_scalar
{
  my ($self,$pkg,$field) = @_;
  my $value = $self->get_sync_db_field($pkg,$field);
  return (ref($value) eq 'ARRAY') ? $value->[0] : $value;
}

sub get_pkgs
{
  my ($self) = @_;
  return @{ $self->{PKGS} };
}

sub get_repo
{
  my ($self,$pkg) = @_;
  if ($pkg =~ m/^([^\/]+)\/[^\/]+$/)
  {
    return $1;
  }
#  $pkg = $self->strip_ver_spec($pkg);
  $self->load_sync_db if not defined($self->{DB}->{SYNC});
  return if not $pkg or not defined($self->{DB}->{REPOMAP}->{$pkg});
  return $self->{DB}->{REPOMAP}->{$pkg};
}

sub get_repo_data
{
  my ($self,$pkg) = @_;
  return $self->{DB}->{SYNC}->{$pkg} if defined($self->{DB}->{SYNC}->{$pkg});
  return {};
}

sub get_repo_servers
{
  my ($self,$repo) = @_;
  return $self->{PacmanConf}->get_servers($repo);
}

sub get_sync_db_ver
{
  my ($self,$pkg) = @_;
  return $self->{DB}->{SYNC}->{$pkg}->{VERSION} if defined($self->{DB}->{SYNC}->{$pkg});
}

sub get_sync_db_filename
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_scalar($pkg,'FILENAME');
}

sub get_sync_db_filtered_deps
{
  my ($self,$pkg) = @_;
  my @deps = $self->get_sync_db_deps($pkg);
  return map {$self->strip_ver_spec($_)} @deps;
}

sub get_sync_db_group_pkgs
{
  my ($self,$group) = @_;
  $self->load_sync_db if not defined($self->{DB}->{SYNC});
  my %group_pkgs;
  foreach my $pkg ($self->get_sync_db_pkgs)
  {
    $group_pkgs{$pkg} ++ if ($self->is_in_sync_db_group($pkg,$group));
  }
  return keys %group_pkgs;
}

sub get_sync_db_groups
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_array($pkg,'GROUPS');
}

sub get_sync_db_isize
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_scalar($pkg,'ISIZE');
}

sub get_sync_db_md5sum
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_scalar($pkg,'MD5SUM');
}

sub get_sync_db_missing_deps
{
  my ($self,$pkg,@missing_deps) = @_;
  @missing_deps = () if not (@missing_deps);
  foreach my $dep ($self->get_sync_db_deps($pkg))
  {
    if (not (grep {$_ eq $dep} @missing_deps or $self->is_in_local_db($dep) ))
    {
      push @missing_deps,$dep;
      push @missing_deps,$self->get_sync_db_missing_deps($dep,@missing_deps)
    }
  }
}

sub get_sync_db_pkgs
{
  my ($self) = @_;
  my @pkgs = keys %{ $self->{DB}->{SYNC} };
  my @ordered_pkgs = ();
  foreach my $repo ($self->{PacmanConf}->get_repos())
  {
    my $l = length($repo) + 1;
    push @ordered_pkgs, sort grep {substr($_,0,$l) eq $repo.'/'} @pkgs;
  }
  return @ordered_pkgs;
}

sub get_sync_db_provider
{
  my ($self,$name) = @_;
  $self->load_sync_db if not defined($self->{DB}->{SYNC});
  foreach my $pkg ($self->get_sync_db_pkgs)
  {
    return $pkg if ($self->is_in_sync_db_provides($pkg,$name));
  }
  return undef;
}

sub get_sync_db_provides
{
  my ($self,$pkg) = @_;
  return $self->get_sync_db_field_as_array($pkg,'PROVIDES');
}

sub get_sync_first
{
  my ($self) = @_;
  return $self->{PacmanConf}->get_option('SyncFirst');
}


sub get_upgradable_pkgs_from_binary
{
  my ($self) = @_;
  my $cmd = $self->get_binary;
  $cmd .= ' -Quq';
  my @pkgs = map {chomp $_; $_} $self->{Shell}->execute($cmd);
  return @pkgs;
}

sub get_sysupgrade_download_from_binary
{
  my ($self,$download) = @_;
  # the ignore arrays are needed to skip ignored in pacman's "-Qu" output
  my @ignored_pkgs = $self->get_ignored_pkgs;
  my @ignored_groups = $self->get_ignored_groups;
  foreach my $name ($self->get_upgradable_pkgs_from_binary)
  {
    my $pkg = $self->maybe_prepend_repo($name);
    next if grep {$name eq $_} @ignored_pkgs;
    next if grep {$self->is_in_sync_db_group($pkg,$_)} @ignored_groups;
    $download = $self->add_to_download($pkg,$download);
  }
  return $download;
}

sub get_sysupgrade_download
{
  my ($self,$download) = @_;
  $self->load_local_db if not defined($self->{DB}->{LOCAL});
  $self->load_sync_db if not defined($self->{DB}->{SYNC});
  my @ignored_pkgs = $self->get_ignored_pkgs;
  my @ignored_groups = $self->get_ignored_groups;
  foreach my $name ($self->get_local_db_pkgs)
  {
    #my $name = $self->get_name( $self->strip_ver_spec($pkg) );
    my $pkg = $self->maybe_prepend_repo($name);
    next if not $self->is_in_sync_db($pkg);
    next if grep {$name eq $_} @ignored_pkgs;
    next if grep {$self->is_in_sync_db_group($pkg,$_)} @ignored_groups;
    my $local_ver = $self->get_local_db_ver($name);
    my $repo_ver = $self->get_sync_db_ver($pkg);
    my $cmp_ver = &compare_versions($repo_ver,$local_ver);
    if ($cmp_ver == 1 or ($cmp_ver == -1 and $self->is_forced($pkg)))
    {
      $download = $self->add_to_download($pkg,$download);
    }
  }
  return $download;
}

sub get_urls
{
  my ($self,$pkg) = @_;
  my $repo = $self->get_repo($pkg);
  my @repo_servers = $self->get_repo_servers($repo);
  my $filename = $self->get_sync_db_filename($pkg);
  return map {$_.'/'.$filename} @repo_servers;
}

sub is_forced
{
  my ($self,$pkg) = @_;
  return defined($self->get_sync_db_field_as_scalar($pkg,'FORCE'));
}

sub is_in_local_db
{
  my ($self,$pkg) = @_;
  $pkg = $self->maybe_strip_repo($pkg);
  my ($name,$min_ver) = $self->get_name_and_min_ver($pkg);
  my $pkg_data = $self->get_local_db_data($name);
  if (
    defined($pkg_data)
  and
    &compare_versions( $pkg_data->{VERSION},$min_ver ) > -1
  )
  {
    return 1;
  }
  elsif ( $self->get_local_db_provider($pkg) )
  {
    return 1;
  }
  else
  {
    return 0;
  }
}

sub is_in_local_db_provides
{
  my ($self,$pkg,$name) = @_;
  my @provides = $self->get_local_db_provides($pkg);
  return $self->is_in_provides($name,@provides);
}

sub is_in_provides
{
  my ($self,$required,@provides) = @_;
  $required = $self->maybe_strip_repo($required);
  my ($required_name,$required_ver) = $self->get_name_and_min_ver($required);
  foreach my $provided (@provides)
  {
    my ($provided_name,$provided_ver) = $self->get_name_and_min_ver($provided);
    next if not $required_name eq $provided_name;
    return 1 if &compare_versions($provided_ver,$required_ver) > -1;
  }
  return 0;
}

sub is_in_sync_db
{
  my ($self,$pkg) = @_;
  my ($name,$min_ver) = $self->get_name_and_min_ver($pkg);
  $name = $self->maybe_prepend_repo($name);
  my $pkg_data = $self->get_sync_db_data($name);
  if (
    defined($pkg_data)
  and
    &compare_versions( $pkg_data->{VERSION},$min_ver ) > -1
  )
  {
    return $name;
  }
  else
  {
    return 0;
  }
}

sub is_in_sync_db_group
{
  my ($self,$pkg,$group) = @_;
  my @groups = $self->get_sync_db_groups($pkg);
  return (@groups and grep {$_ eq $group} @groups) ? 1 : 0;
}

sub is_in_sync_db_provides
{
  my ($self,$pkg,$name) = @_;
  my @provides = $self->get_sync_db_provides($pkg);
  return $self->is_in_provides($name,@provides);
}

sub is_possible_download
{
  my ($self) = @_;
  return 0 if $self->{OPERATION} ne 'SYNC';
  foreach my $arg ($self->get_args)
  {
    return 0 if grep {$_ eq $arg} @NODOWNLOAD;
  }
  return 0 if scalar($self->get_pkgs) eq 0 and not $self->get_arg('--sysupgrade');
  return 1;
}

sub load_db_file
{
  my ($self,$path,$pkg_data) = @_;
  my $file_data = &parse_db_file($path);
  return {%$pkg_data, %$file_data};
}

sub load_local_db
{
  my ($self) = @_;
  $self->{DB}->{LOCAL} = {};
  my $db_dir = $self->{PacmanConf}->get_option('DBPath').'/local';
  return if not -d $db_dir;
  opendir(my $dh,$db_dir);
  foreach my $item (readdir($dh))
  {
    if ($item =~ m/^(.+)-(.+-.+)$/)
    {
      $self->{DB}->{LOCAL}->{$1}->{VERSION} = $2;
    }
  }
  close $dh;
}

sub load_local_db_file
{
  my ($self,$pkg,$file) = @_;
  my $pkg_data = $self->{DB}->{LOCAL}->{$pkg};
  if (not defined($pkg_data->{"[$file]"}))
  {
    my $ver = $pkg_data->{VERSION};
    my $file_path = $self->{PacmanConf}->get_option('DBPath')."/local/$pkg-$ver/$file";
    $pkg_data = $self->load_db_file($file_path,$pkg_data);
    $pkg_data->{"[$file]"} = 1;
    $self->{DB}->{LOCAL}->{$pkg} = $pkg_data;
  }
}

sub load_sync_db
{
  my ($self) = @_;
  $self->{DB}->{SYNC} = {};
  $self->{DB}->{REPOMAP} = {};
  my $db_dir = $self->{PacmanConf}->get_option('DBPath').'/sync';
  return if not -d $db_dir;
  foreach my $repo ($self->{PacmanConf}->get_repos)
  {
    my $repo_dir = $db_dir.'/'.$repo;
    next if not -d $repo_dir;
    opendir(my $dh,$repo_dir);
    foreach my $item (readdir($dh))
    {
      if ($item =~ m/^(.+)-(.+-.+)$/)
      {
        $self->{DB}->{SYNC}->{"$repo/$1"}->{VERSION} = $2;
        $self->{DB}->{REPOMAP}->{$1} = $repo if not defined($self->{DB}->{REPOMAP}->{$1});
      }
    }
    close $dh;
  }
}

sub load_sync_db_file
{
  my ($self,$pkg,$file) = @_;
  my $pkg_data = $self->{DB}->{SYNC}->{$pkg};
  if (not defined($pkg_data->{"[$file]"}))
  {
    my $ver = $pkg_data->{VERSION};
    my $file_path = $self->{PacmanConf}->get_option('DBPath')."/sync/$pkg-$ver/$file";
    $pkg_data = $self->load_db_file($file_path,$pkg_data);
    $pkg_data->{"[$file]"} = 1;
    $self->{DB}->{SYNC}->{$pkg} = $pkg_data;
  }
}

sub maybe_prepend_repo
{
  my ($self,$pkg) = @_;
  if ($pkg !~ m/^[^\/]+\/[^\/]+$/)
  {
    my $repo = $self->get_repo($pkg);
    $pkg = "$repo/$pkg" if defined $repo;
  }
  return $pkg;
}

sub maybe_refresh_only
{
  my ($self) = @_;
  if (defined($self->{ARGS}->{'--refresh'}) )
  {
    $self->refresh_only;
  }
}

sub maybe_strip_repo
{
  my ($self,$pkg) = @_;
  if ($pkg =~ m/^[^\/]+\/([^\/]+)$/)
  {
    $pkg = $1;
  }
  return $pkg;
}

sub parse_args
{
  my ($self,@args) = @_;
  @args = &dejuxtapose_args(@args);
  @args = &hash_map($arg_maps->{EXPANSIONS}->{COMMON}, @args);
  foreach my $arg (@args)
  {
    if (defined($arg_maps->{OPERATIONS}->{$arg}))
    {
      my $operation = $arg_maps->{OPERATIONS}->{$arg};
      if ($self->{OPERATION} eq 'NONE')
      {
        $self->{OPERATION} = $operation;
      }
      elsif ($self->{OPERATION} ne 'ERROR')
      {
        if ($operation eq 'HELP') {$self->{OPERATION} = 'HELP';}
        elsif ($operation eq 'VERSION' and $self->{OPERATION} ne 'HELP') {$self->{OPERATION} = 'VERSION';}
        elsif ($operation ne 'VERSION') {$self->{OPERATION} = 'ERROR';}
      }
    }
  }
  my $expansions;
  if (defined($arg_maps->{EXPANSIONS}->{ $self->{OPERATION} }))
  {
    #combine the common expansion map with the oen for the specified operation, if there is one
    $expansions = {%{ $arg_maps->{EXPANSIONS}->{COMMON} }, %{ $arg_maps->{EXPANSIONS}->{ $self->{OPERATION} } } };
  }
  else
  {
    $expansions = {%{ $arg_maps->{EXPANSIONS}->{COMMON} }}
  }
  $self->{EXPANSIONS} = $expansions;
  # expand the arguments
  @args = &hash_map($expansions,@args);
  $self->{ARGS}->{ARRAY} = [@args];
  my %reset;

  while (@args)
  {
    my $arg = shift @args;
    if (defined($arg_maps->{PARAMETRIC}->{$arg}))
    {
      my $parameter = @args ? shift @args : '';
      if (defined($arg_maps->{CONF}->{$arg}))
      {
        my $option = $arg_maps->{CONF}->{$arg};
        $self->{PacmanConf} = $self->{PacmanConf}->set_option_by_arg($option,$parameter);
      }
      else
      {
        $self->{ARGS}->{$arg} = $parameter;
      }
    }
    elsif (substr($arg,0,1) eq '-')
    {
      $self->{ARGS}->{$arg} ++;
    }
    else
    {
      $self->add_pkg($arg);
    }
  }
  return $self->{OPERATION};
}

sub refresh_only
{
  my ($self) = @_;
  $self->set_callback(sub{$|++;print @_;});
  $self->execute($REFRESH_FILTER);
  $self->set_callback(undef);
  $self->clear_arg('--refresh');
}

sub requires_root
{
  my ($self,@args) = @_;
  foreach my $arg (@args)
  {
    return 1 if grep {$_ eq $arg} @NEEDSROOT;
  }
  return 0;
}

sub set_callback
{
  my ($self,$callback) = @_;
  $self->{Shell}->{CALLBACK} = $callback;
}

sub set_binary
{
  my ($self,$binary) = @_;
  $self->{BINARY} = $binary;
}

sub strip_ver_from_name
{
  my ($self,$pkg) = @_;
  $pkg =~ s/-[^-]+-[\d\.]*\d$//;
  return $pkg;
}

sub strip_ver_spec
{
  my ($self,$pkg) = @_;
  $pkg =~ s/(?:>|=|>=).+$//;
  return $pkg;
}

sub warning
{
  my ($self,@output) = @_;
  print STDERR @output;
}

################################
########## CLASS SUBS ##########
################################

sub compare_versions
{
  my ($a,$b) = @_;
  return -1 if not defined($a) and defined($b);
  return 1 if defined($a) and not defined($b);
  return 0 if not defined($a) and not defined($b);

  my ($rel_a,$rel_b) = (0,0);
  if ($a =~ m/(.*)-(.+)$/) {($a,$rel_a)=($1,$2);}
  if ($b =~ m/(.*)-(.+)$/) {($b,$rel_b)=($1,$2);}

  my @a = split(/\./,$a);
  my @b = split(/\./,$b);
  my $len_a = scalar @a;
  my $len_b = scalar @b;
  my $n = ($len_a < $len_b) ? $len_a : $len_b;
  for (my $i=0; $i<$n; $i++)
  {
    my ($int_a,$alph_a,$dash_a) = ($a[$i] =~ m/^(\d*)(.*?)(_.*)?$/);
    my ($int_b,$alph_b,$dash_b) = ($b[$i] =~ m/^(\d*)(.*?)(_.*)?$/);
    $int_a = 0 if $int_a eq '';
    $int_b = 0 if $int_b eq '';
    if ($int_a != $int_b)
    {
      return $int_a <=> $int_b;
    }
    if (length($alph_a) and not length($alph_b))
    {
      return -1;
    }
    if (not length($alph_a) and length($alph_b))
    {
      return 1;
    }
    if ($alph_a ne $alph_b)
    {
      return $alph_a cmp $alph_b;
    }
    return 1 if defined($dash_a) and not defined($dash_b);
    return -1 if not defined($dash_a) and defined($dash_b);
    if (defined($dash_a) and defined($dash_b))
    {
      my $dash_cmp = $dash_a cmp $dash_b;
      return $dash_cmp if $dash_cmp != 0;
    }
  }
  return $len_a <=> $len_b if not $len_a == $len_b;
  return $rel_a <=> $rel_b;
}

sub parse_db_file
{
  my ($file) = @_;
  my $data = {};
  my $key;
  my @values = ();
  my $file_contents;
  open(my $fh,'<',$file) or die "failed to open $file: $!\n";
  {
    local $/;
    $file_contents = <$fh>;
  }
  close $fh;
  foreach my $line (split "\n", $file_contents)
  {
    if ($line =~ m/^\s*%([^%]+)%\s*$/)
    {
      if (defined($key) and @values)
      {
        if (scalar @values > 1)
        {
          $data->{$key} = [@values];
        }
        else
        {
          $data->{$key} = $values[0];
        }
        @values = ();
      }
      elsif (defined($key))
      {
        $data->{$key} = '';
      }
      $key = $1;
      next;
    }
    next if (not defined($key) or ($line =~ m/^\s*$/));
    $line =~ s/^\s*//g;
    $line =~ s/\s*$//g;
    push @values, $line;
  }

  if (defined($key) and @values)
  {
    if (scalar @values > 1)
    {
      $data->{$key} = [@values];
    }
    else
    {
      $data->{$key} = $values[0];
    }
  }
  elsif (defined($key))
  {
    $data->{$key} = '';
  }
  return $data;
}

1;

__END__

=head1 NAME

Xyne::Arch::Pacman - A pacman class.

=head1 SYNOPSIS

  use Xyne::Arch::Pacman;

=head1 DESCRIPTION

A partial implementation of pacman primarily for use in
scripts. The pacman binary is still used for more functionality.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
