# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::PacmanConf;
use strict;
use warnings;

use Xyne::Common::Functions qw/strip_trailing_slash/;

our $VERSION = '0.04';

my $OPTIONS = {
  'CacheDir'      => 'list',
  'CleanMethod'   => 'str',
  'DBPath'        => 'str',
  'HoldPkg'       => 'list',
  'IgnoreGroup'   => 'list',
  'IgnorePkg'     => 'list',
  'Include'       => 'str',
  'LogFile'       => 'str',
  'NoExtract'     => 'list',
  'NoPassiveFtp'  => 'switch',
  'NoUpgrade'     => 'list',
  'RootDir'       => 'str',
  'Server'        => 'str',
  'ShowSize'      => 'switch',
  'SyncFirst'     => 'list',
  'TotalDownload' => 'switch',
  'UseDelta'      => 'switch',
  'UseSyslog'     => 'switch',
  'XferCommand'   => 'str'
};

my @PATHS = qw/CacheDir DBPath LogFile RootDir/;
my @RELATIVE_PATHS = qw/DBPath LogFile/;
my @LISTS = grep {$OPTIONS->{$_} eq 'list'} keys %${OPTIONS};
my @SWITCHES = grep {$OPTIONS->{$_} eq 'switch'} keys %${OPTIONS};


my $DEFAULTS = {
  'CacheDir' => '/var/cache/pacman/pkg',
  'DBPath'   => '/var/lib/pacman',
  'LogFile'  => '/var/log/pacman.log',
  'RootDir'  => '',
  
  # Module Defaults
  'Config'   => '/etc/pacman.conf'
};

sub new {
  my ($proto,$args) = @_;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  $self->{PATH} = $args->{PATH} ? $args->{PATH} : $DEFAULTS->{Config};
  return $self;
}







sub append_servers
{
  my ($self,$repo,@servers) = @_;
  return if not defined($repo);
  foreach my $server (@servers)
  {
    $server = &strip_trailing_slash($server);
    $server =~ s/\$repo(?!\w)/$repo/g;
    push @{ $self->{repos}->{$repo} }, $server if not grep {$_ eq $server} @{ $self->{repos}->{$repo} };
  }
}

sub clear_all
{
  my ($self,$option) = @_;
  $self->clear_option($_) foreach keys %{ $self->{options} };
  $self->clear_repo($_) foreach $self->get_repos;
}

sub clear_option
{
  my ($self,$option) = @_;
  return if not defined($OPTIONS->{$option});
  delete($OPTIONS->{$option});
}

sub clear_repo
{
  my ($self,$repo) = @_;
  delete($self->{repos}->{$repo});
}

sub error
{
  my ($self,$error) = @_;
  print STDERR "$error";
  exit 1;
}

sub get_config
{
  my ($self) = @_;
  my $config = "[options]\n";
  foreach my $option (keys %{ $self->{options} })
  {
    next if not defined($OPTIONS->{$option});
    if ($OPTIONS->{$option} eq 'switch')
    {
      $config .= "$option\n";
    }
    elsif ($OPTIONS->{$option} eq 'list')
    {
      my @values = @{ $self->{options}->{$option} };
      if ($option ne 'CacheDir')
      {
        $config .= $option.' = '.join(' ',@values)."\n";
      }
      else
      {
        $config .= "$option = $_\n" foreach @values;
      }
    }
    else
    {
      $config .= "$option = $self->{options}\n";
    }
  }
  foreach my $repo (@{ $self->{repo_list} })
  {
    $config .= "\n[$repo]\n";
    $config .= "Server = $_\n" foreach $self->get_servers($repo);
  }
  return $config;
}

sub get_option
{
  my ($self,$option) = @_;
  return if not defined($OPTIONS->{$option});
  if (defined($self->{options}->{$option}))
  {
    if ($OPTIONS->{$option} eq 'list')
    {
      return @{ $self->{options}->{$option} };
    }
    else
    {
      return $self->{options}->{$option};
    }
  }
  elsif (grep {$_ eq $option} @RELATIVE_PATHS)
  {
    return $self->get_option('RootDir').$DEFAULTS->{$option};
  }
  elsif (defined($DEFAULTS->{$option}))
  {
    return $DEFAULTS->{$option};
  }
  elsif ($OPTIONS->{$option} eq 'switch')
  {
    return 0;
  }
  return;
}

sub get_option_set_by_arg
{
  my ($self,$option) = @_;
  return $self->get_option($option) if defined($self->{args}->{$option});
}

sub get_path
{
  my ($self) = @_;
  return $self->{PATH};
}

sub get_repos
{
  my ($self) = @_;
  return @{ $self->{repo_list} };
}

sub get_servers
{
  my ($self,$repo) = @_;
  return () if not defined($repo) or not defined($self->{repos}->{$repo});
  return grep {$_} @{ $self->{repos}->{$repo} };
}

sub load_config
{
  my ($self,$args,@loaded) = @_;
  my $file_path = defined($args->{'file_path'}) ? $args->{'file_path'} : $self->{PATH};
  # Return if this file has been loaded already. This avoids cyclic recursion.
  return $self if (grep {$_ eq $file_path} @loaded);
  my $repo = $args->{'repo'} if defined($args->{'repo'});

  $self->error("error: invalid pacman configuration file specified") if not $file_path;
  $self->error("error: $file_path does not exist") if not -e $file_path;
  open (my $fh, '<', $file_path) or $self->error("error: failed to open $file_path ($!)");
  while ( defined(my $line = <$fh>) )
  {
    next if ($line =~ m/^\s*#/);
    if ($line =~ m/^\s*\[([^\s\]]+)\]\s*$/)
    {
      $repo = $1;
      push @{ $self->{repo_list} }, $repo if $repo ne 'options';
      $args->{'repo'} = $repo;
    }
    if ($line =~ m/^\s*(\S+)\s*=\s*(.+)\s*$/)
    {
      if ($repo eq 'options')
      {
        $self->set_option($1,$2);
      }
      elsif ($1 eq 'Include')
      {
        $args->{'file_path'} = $2;
        $self->load_config($args,@loaded,$file_path);
      }
      elsif ($1 eq 'Server' and $repo)
      {
        $self->append_servers($repo,$2);
      }
    }

    elsif ($line =~ m/^\s*(\S+)\s*$/ and $repo eq 'options')
    {
      $self->set_option($1);
    }
  }
  close $fh;
  return $self;
}

sub load_config_by_arg
{
  my ($self,$path) = @_;
  return if not defined($path);
  my $config = $self->new({
    PATH => $path
  });
  $config->load_config;
  foreach my $option (keys %{ $self->{args} })
  {
    $config->set_option_by_arg($option, $self->get_option_set_by_arg($option));
  }
  $self = $config;
  return $config;
}

sub reload
{
  my ($self) = @_;
  $self->clear_all;
  $self->load_config;
}

sub save_config
{
  my ($self,$path) = @_;
  open(my $fh, '>', $path) or $self->error("error: unable to open $path ($!)");
  print $fh $self->get_config;
  close $fh;
}

sub set_path
{
  my ($self,$path) = @_;
  $self->{PATH} = $path if defined($path);
}

sub set_option
{
  my ($self,$option,$value) = @_;
  return 0 if not defined($OPTIONS->{$option});
  my $type = $OPTIONS->{$option};
  $value = &strip_trailing_slash($value) if (grep {$_ eq $option} @PATHS);
  if ($type eq 'switch')
  {
    if (defined($value) and not $value)
    {
      delete($self->{options}->{$option});
    }
    else
    {
      $self->{options}->{$option} = 1;
    }
  }
  elsif ($type eq 'list')
  {
    if ($option eq 'CacheDir' and defined($value))
    {
      push @{ $self->{options}->{$option} }, $value;
    }
    else
    {
      push @{ $self->{options}->{$option} }, grep {$_} split(/\s+/,$value);
    }
  }
  else
  {
    $self->{options}->{$option} = $value if defined($value);
  }
  return 1;
}

sub set_option_by_arg
{
  my ($self,$option,$value) = @_;
  if ($option eq 'Config')
  {
    return $self->load_config_by_arg($value);
  }
  $self->{args}->{$option} = 1 if $self->set_option($option,$value);
  return $self;
}

1;

__END__

=head1 NAME

Xyne::Arch::PacmanConf - Pacman configuration file class.

=head1 SYNOPSIS

  use Xyne::Arch::PacmanConf;

=head1 DESCRIPTION

A class to load and parse pacman configuration files and
store pacman settings, including arguments passed to it from
pacman objects.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
