# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Rebase;
use strict;
use warnings;

use IPC::Open2;
use LWP::Simple qw/get head/;
use POSIX qw/uname/;
use Cwd qw/abs_path/;

use Xyne::Common::Functions qw/colorize decolorize find_all mkdir_p strip_trailing_slash/;
use Xyne::Common::Log;
use Xyne::Arch::PacmanConf;

our $VERSION = '0.16';

my $OPTIONS = {
  '--arch'             => 'parametric',
  '--by-lastupdate'    => 'switch',
  '--dir'              => 'parametric',
  '--full-extract'     => 'switch',
  '--get-tarballs'     => 'switch',
  '--help'             => 'switch',
  '--log'              => 'parametric',
  '--no-color'         => 'switch',
  '--no-sync-dir'      => 'switch',
  '--pacman-config'    => 'parametric',
  '--reset'            => 'switch',
  '--silent'           => 'switch'
};

my @PATHS = qw/--dir --log --pacman-config/;

my $DEFAULTS = {
  'SubDir'     => 'sync',
  'LockFile'   => 'db.lck',
  'DBExt'      => 'db.tar.gz',
  'lastupdate' => '.lastupdate'
};

my $EXPANSIONS = {
  '-h' => '--help',
};

my @ARCHITECTURES = qw/i686 x86_64/;
my $TAR_BIN;
my @TAR_BINS = qw/bsdtar tar/;
while (@TAR_BINS)
{
  my $bin = shift @TAR_BINS;
  $bin = `which $bin 2>/dev/null`;
  chomp $bin;
  if (defined($bin) and -x $bin)
  {
    $TAR_BIN = $bin;
    last;
  }
}

my $DEFAULT_WIDTH;
if (-x '/bin/tput')
{
  $DEFAULT_WIDTH = `/bin/tput cols`;
}
else
{
  $DEFAULT_WIDTH = 79;
}


sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  $self->{ARCH} = [uname()]->[4];
  my $args = shift;
  if (defined($args->{PacmanConf}))
  {
    $self->{PacmanConf} = $args->{PacmanConf};
  }
  else
  {
    $self->{PacmanConf} = Xyne::Arch::PacmanConf->new();
  }
  $self->{PacmanConf}->load_config;
  $self->set_option('--dir',$self->{PacmanConf}->get_option('DBPath'));

  $self->set_display_width($DEFAULT_WIDTH);
  
  $self->{'DB_LOCKED'} = 0;

  return $self;
}





sub append_repo
{
  my ($self,$repo) = @_;
  push @{ $self->{REPOS} }, $repo;
}

sub append_tag
{
  my ($self,$str,$tag,$color) = @_;
  my $width = $self->get_display_width;
  my $tag_width = length($tag)+2;
  $tag = &colorize('[',$color).$tag.&colorize(']',$color);
  my $n = $width - length(&decolorize($str)) - $tag_width ;
  return $str.(' ' x $n).$tag
}

sub display_help
{
  print <<'HELP';
USAGE
  rebase [options] [repos]

OPTIONS
  --arch <architecture>
  --by-lastupdate
  --dir /path/to/dir
  --full-extract
  --get-tarballs
  --help
  --log /path/to/file
  --no-color
  --no-sync-dir
  --pacman-config /path/to/pacman.conf
  --reset
  --silent

See the man page for details.

HELP
}

sub extract_db
{
  my ($self,$repo,$check_lastupdate) = @_;
  my $output_dir = $self->get_output_dir;
  my $repo_dir = "$output_dir/$repo";
  my $lastupdate;
  if ($check_lastupdate)
  {
    $lastupdate = $self->get_lastupdate($repo_dir);
  }
  my ($db_tarball,$url,$remote_lastupdate) = $self->get_database($repo,$lastupdate);
  if (not ( defined($db_tarball) or defined($remote_lastupdate) ))
  {
    $self->msg_warning("failed to retrieve database for $repo\n");
    return;
  }
  if (not defined($db_tarball))
  {
    $self->msg_header('bold blue',$repo, $url);
    $self->msg_synchronized($repo,0,0,0);
    return;
  }
  if (not -e $repo_dir)
  {
    # mkdir_p returns "1" if it fails (to emulate mkdir -p)
    $self->msg_error("unabled to create directory: $repo_dir") if &mkdir_p($repo_dir)
  }
  elsif (not -w $repo_dir)
  {
    $self->msg_error("unable to write to $repo_dir: permission denied\n");
  }

  my @files = split "\n", &run_tar_cmd($db_tarball,'-t');
  my %files = ();
  my @new_pkgs = ();
  foreach my $file (map {&strip_trailing_slash($_)} @files)
  {
    if (-e "$repo_dir/$file")
    {
      $files{$file} = 1;
      next;
    }
    else
    {
      $files{$file} = 2;
      if ($file =~ m/^([^\/]+)$/)
      {
        push @new_pkgs,$1;
      }
    }
  }

  #$self->msg_new($repo,$url,@new_pkgs);
  $self->msg_header('bold blue',$repo, $url);

  &mkdir_p($repo_dir);
  my @filelist;
  if ($self->get_option('--full-extract'))
  {
    $self->msg_normal("performing a full extraction\n");
    &run_tar_cmd($db_tarball,"-C $repo_dir -x 2>/dev/null");
    @filelist = keys %files;
  }
  else
  {
    &run_tar_cmd($db_tarball,"-C $repo_dir -xk 2>/dev/null");
    @filelist = grep {$files{$_}==2} keys %files;
  }
  #my @added;
  foreach my $file (@filelist)
  {
    if (not -e "$repo_dir/$file")
    {
      $self->msg_error("failed to create $file in $repo\n");
    }
    #else
    #{
    #  push @added, $file;
    #}
  }

  #$self->msg_added($repo,sort @added);

  #my @removed = ();
  my %removed = ();
  my @local_files = &find_all($repo_dir);
  foreach my $path (@local_files)
  {
    my $file = substr($path,length($repo_dir)+1);
    next if substr($file,0,1) eq '.';
    if (not defined( $files{$file} ) )
    {
      if (-d $path) {rmdir $path;}
      else {unlink $path;}
      if (-e $path)
      {
        $self->msg_warning("failed to remove $path\n");
      }
      else
      {
        #push @removed, $file;
        if ($file =~ m/^([^\/]+)$/)
        {
          my ($pkgname,$pkgver) = &get_pkgname_and_pkgver($file);
          $removed{$pkgname} = $pkgver;
        }
      }
    }
  }

  #$self->msg_removed($repo,sort @removed);
  my $new_pkgs = 0;
  my $upgraded_pkgs = 0;
  my $old_pkgs = 0;
  foreach my $pkg (sort @new_pkgs)
  {
    my ($pkgname,$pkgver) = &get_pkgname_and_pkgver($pkg);
    if (exists($removed{$pkgname}))
    {
      $self->msg_upgraded($repo,$pkgname,$removed{$pkgname},$pkgver);
      delete($removed{$pkgname});
      $upgraded_pkgs++;
    }
    else
    {
      $self->msg_new($repo,$pkgname,$pkgver);
      $new_pkgs++;
    }
  }

  foreach my $pkg (sort keys %removed)
  {
    $self->msg_old($repo,$pkg,$removed{$pkg});
    $old_pkgs++;
  }


  $self->msg_synchronized($repo,$new_pkgs,$upgraded_pkgs,$old_pkgs);
  
  if ($check_lastupdate and $remote_lastupdate)
  {
    $self->update_lastupdate($repo_dir,$remote_lastupdate);
  }
}

sub get_database
{
  my ($self,$repo,$lastupdate) = @_;
  return if not defined($repo);
  my @servers = $self->{PacmanConf}->get_servers($repo);
  if (not @servers)
  {
    $self->msg_error("unknown repo: $repo\n");
    return undef;
  }
  foreach my $server (@servers)
  {
    my $url = "${server}/${repo}.$DEFAULTS->{DBExt}";
    my $target_arch = $self->get_option('--arch');
    if (defined($target_arch))
    {
      foreach my $arch (@ARCHITECTURES)
      {
        next if $arch eq $target_arch;
        $url =~ s/\Q$arch\E/$target_arch/g;
      }
    }
    my $remote_lastupdate;
    if (defined($lastupdate))
    {
      $remote_lastupdate = [head $url]->[2];
      if (not defined($remote_lastupdate))
      {
        return (undef, undef, undef);
      }
      elsif (not $remote_lastupdate > $lastupdate)
      {
        return (undef, $url, $remote_lastupdate);
      }
    }
    my $db_tarball = get($url);
    return ($db_tarball,$url,$remote_lastupdate) if defined($db_tarball);
    $self->msg_warning("failed to retrieve $url\n");
  }
  return (undef, undef, undef);
}

sub get_lastupdate
{
  my ($self,$dir) = @_;
  my $fpath = $dir . '/' . $DEFAULTS->{'lastupdate'};
  if (-r $fpath)
  {
    open(my $fh, '<', $fpath) or $self->msg_error("failed to open $fpath\n");
    my $lastupdate = <$fh>;
    close $fh;
    chomp $lastupdate if $lastupdate;
    $lastupdate = 0 if not $lastupdate;
    return $lastupdate;
  }
  return 0;
}

sub update_lastupdate
{
  my ($self,$dir,$lastupdate) = @_;
  my $fpath = $dir . '/' . $DEFAULTS->{'lastupdate'};
  open(my $fh, '>', $fpath) or $self->msg_error("failed to open $fpath\n");
  print $fh $lastupdate;
  close $fh;
}

sub get_option
{
  my ($self,$option) = @_;
  return if not defined($option) or not defined($OPTIONS->{$option});
  return $self->{ARGS}->{$option} if defined($self->{ARGS}->{$option});
  return 0 if $OPTIONS->{$option} eq 'switch';
  return undef;
}

sub get_output_dir
{
  my ($self) = @_;
  my $dir;
  if ($self->get_option('--get-tarballs'))
  {
    $dir = '.';
  }
  else
  {
    $dir = &strip_trailing_slash($self->get_option('--dir'));
    $dir .= '/'.$DEFAULTS->{'SubDir'} unless $self->get_option('--no-sync-dir');
  }
  return &abs_path($dir);
}

sub get_repos
{
  my ($self) = @_;
  return @{ $self->{REPOS} } if defined($self->{REPOS});
  return $self->{PacmanConf}->get_repos;
}

sub get_display_width
{
  my ($self) = @_;
  return $self->{DISPLAY_WIDTH};
}

sub lock_db
{
  my ($self) = @_;
  my $dir = $self->get_option('--dir');
  $self->msg_error("unable to write to $dir: permission denied\n") if not -w $dir;
  my $lock_file = "$dir/$DEFAULTS->{LockFile}";
  $self->msg_normal("locking database ($lock_file)\n");
  $self->msg_error("unable to lock database: a lock file already exists ($lock_file)\n") if -e $lock_file;
  open (my $fh,'>',$lock_file) or return 0;
  close $fh;
  $self->{'DB_LOCKED'} = 1;
  return 1;
}

sub maybe_colorize
{
  my ($self,$str,$color) = @_;
  if ($self->get_option('--no-color'))
  {
    return $str;
  }
  else
  {
    return &colorize($str,$color);
  }
}

sub msg
{
  my ($self,$msg,$args) = @_;
  my $nocolor = $self->get_option('--no-color');
  my $tag = ($args->{'tag'}) ? $args->{'tag'} : '';
  if (not $self->get_option('--silent'))
  {
    if ($tag ne '' and not $nocolor and $args->{'color'})
    {
      print &colorize($tag,$args->{'color'}),$msg;
    }
    else
    {
      print $tag,$msg;
    }
  }
  if (defined($self->{LOG}) or $self->get_option('--log'))
  {
    $self->{LOG} = Xyne::Common::Log->new( $self->get_option('--log') ) if not defined($self->{LOG});
    $self->{LOG}->log($tag.$msg);
  }
}

sub msg_added
{
  my $self = shift;
  $self->msg_changed('+ ', 'bold green', @_);
}

sub msg_changed
{
  my ($self,$tag,$color,$repo,@files) = @_;
  my $note = '';
  my $width = $self->get_display_width;
  $self->set_display_width($width - length($tag) + 1);
  @files = map {$self->append_tag($_,$repo,'bold blue')} @files;
  $self->set_display_width($width);
  foreach my $file (@files)
  {
    $self->msg("$file\n",{
     'tag' => $tag,
      'color' => $color
    });
  }
}

sub msg_error
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '==> ERROR ',
    'color' => 'bold red'
  });
  $self->unlock_db() if $self->{'DB_LOCKED'};
  exit(1);
}

sub msg_header
{
  my ($self,$color,$title,@lines) = @_;
  my $n = $self->get_display_width;
  my $filler = '-';
  foreach my $line (@_)
  {
    my $l = length($line)+4;
    $n = $l if $l > $n;
  }
  my $msg = $self->maybe_colorize(($filler x 1).'[', $color);
  $msg .= $title;
  $msg .= $self->maybe_colorize(']'.($filler x ($n - length($title) - 3) )."\n", $color);
  foreach my $line (@lines)
  {
    $msg .= "  $line\n";
  }
  $msg .= $self->maybe_colorize(($filler x $n),$color)."\n";
  $self->msg($msg);
}

sub msg_new
{
  my ($self,$repo,$name,$ver) = @_;
  my $line = $self->maybe_colorize('+ ','bold green').$name.' '.$self->maybe_colorize($ver,'bold black');
  my $width = $self->get_display_width;
  $line = $self->append_tag($line,$repo,'bold blue');
  $line .= "\n";
  $self->msg($line);
}


sub msg_normal
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '--> ',
    'color' => 'bold blue'
  });
}

sub msg_old
{
  my ($self,$repo,$name,$ver) = @_;
  my $line = $self->maybe_colorize('- ','bold red').$name.' '.$self->maybe_colorize($ver,'bold black');
  my $width = $self->get_display_width;
  $line = $self->append_tag($line,$repo,'bold blue');
  $self->msg($line);
}

sub msg_removed
{
  my $self = shift;
  $self->msg_changed('- ', 'bold red', @_);
}


sub msg_saved
{
  my ($self,$msg) = @_;
  $self->msg("$msg\n",{
   'tag' => '==> ',
    'color' => 'bold green'
  });
}

sub msg_synchronized
{
  my ($self,$repo,$new_pkgs,$upgraded_pkgs,$old_pkgs) = @_;
  my $width = $self->get_display_width;
  my $tag = '==> SYNCHRONIZED ';
  my $i = $width - length($repo) - length($tag) - length($new_pkgs) - length($upgraded_pkgs) - length($old_pkgs) - 7;
  my $spacer = ' ' x $i;
  my $summary = $self->maybe_colorize('+','bold green').$new_pkgs.' '.$self->maybe_colorize('^','bold cyan').$upgraded_pkgs.' '.$self->maybe_colorize('-','bold red').$old_pkgs;
  my $msg = "$repo $spacer $summary\n\n";
  $self->msg($msg,{
   'tag' => $tag,
    'color' => 'bold green'
  });
}

sub msg_upgraded
{
  my ($self,$repo,$name,$oldver,$newver) = @_;
  my $line = $self->maybe_colorize('^ ','bold cyan').$name.' '.$self->maybe_colorize("$oldver -> $newver",'bold black');
  my $width = $self->get_display_width;
  $line = $self->append_tag($line,$repo,'bold blue');
  $self->msg($line);
}

sub msg_warning
{
  my ($self,$msg) = @_;
  $self->msg($msg,{
   'tag' => '==> WARNING ',
    'color' => 'bold red'
  });
}

sub parse_args
{
  my ($self,@args) = @_;
  while (@args)
  {
    my $arg = shift @args;
    next if not defined($arg);
    if ($arg eq '--help')
    {
      $self->display_help;
      exit;
    }
    if (not defined($OPTIONS->{$arg}))
    {
      if (substr($arg,0,1) eq '-')
      {
        $self->msg_warning("ignoring unrecognized option: $arg\n");
      }
      else
      {
        $self->append_repo($arg);
      }
      next;
    }
    if ($OPTIONS->{$arg} eq 'parametric')
    {
      my $parameter = @args ? shift @args : '';
      $parameter = &abs_path($parameter) if (grep {$arg eq $_} @PATHS);
      if ($arg eq '--pacman-config')
      {
        $self->{PacmanConf} = $self->{PacmanConf}->set_option_by_arg('Config',$parameter);
        next;
      }
      $self->{ARGS}->{$arg} = $parameter;
    }
    else
    {
      $self->{ARGS}->{$arg}++;
    }
  }
  $self->_reset if $self->get_option('--reset');
  return $self;
}

sub _reset
{
  my ($self) = @_;
  $self->{PacmanConf}->set_option_by_arg('DBPath', $self->get_output_dir);
  use Xyne::Arch::Pacman;
  my $pacman = Xyne::Arch::Pacman->new({
    'PacmanConf' => $self->{PacmanConf}
  });
  $pacman->parse_args('-Syy');
  my $pacman_bin = $pacman->get_binary;
  $self->msg_normal("running \"$pacman_bin -Syy\" to reset the database\n");
  $pacman->refresh_only;
  exit;
}

sub run
{
  my ($self,@args) = @_;
  $self->parse_args(@ARGV);
  if (not $self->lock_db())
  {
    $self->msg_error("failed to lock database\n");
  }
  else
  {
    foreach my $repo ($self->get_repos)
    {
      if ($self->get_option('--get-tarballs'))
      {
        my ($db_tarball,$url,$remote_lastupdate) = $self->get_database($repo);
        $self->msg_error("failed to retrieve database for $repo\n") if not defined($db_tarball);
        my $file = substr($url,rindex($url,'/'));
        $file = $self->get_output_dir . $file;
        #$self->set_display_width(79);
        $self->msg_header('bold blue', $repo, $url);
        $self->msg_saved("$file\n");
        $self->save_file($file,$db_tarball);
      }
      else
      {
        $self->extract_db($repo, $self->get_option('--by-lastupdate'));
      }
    }
    $self->unlock_db();
  }
}

sub save_file
{
  my ($self,$file,$data) = @_;
  open (my $fh, '>', $file);
  binmode $fh;
  print $fh $data;
  close $fh;
}

sub set_display_width
{
  my ($self,$w) = @_;
  $self->{DISPLAY_WIDTH} = $w;
}

sub set_option
{
  my ($self,$option,$value) = @_;
  return if not defined($option) or not defined($OPTIONS->{$option});
  $self->{ARGS}->{$option} = $value;
}

sub unlock_db
{
  my ($self) = @_;
  my $dir = $self->get_option('--dir');
  my $lock_file = "$dir/$DEFAULTS->{LockFile}";
  $self->msg_normal("unlocking database ($lock_file)\n");
  (-e $lock_file) ? unlink $lock_file : $self->msg_warning("lock file does not exist ($lock_file)\n");
  $self->msg_warning("failed to remove lock file ($lock_file)\n") if -e $lock_file;
}

################################
########## CLASS SUBS ##########
################################

sub get_pkgname_and_pkgver
{
  my ($pkg) = @_;
  if ($pkg =~ m/^(.+)-(.+-.+)$/)
  {
    return ($1,$2)
  }
  return undef;
}

sub run_tar_cmd
{
  umask 0022 if (umask) != 0022;
  local $/;
  if (not defined($TAR_BIN))
  {
    print 'error: this operation requires one of the following: ', join(' ', @TAR_BINS), "\n";
    exit 1;
  }
  my ($input,$options) = @_;
  $options = '' if not defined($options);
  my ($pin,$pout,$pid);
  if (not $pid=open2($pout,$pin,"$TAR_BIN -zf - --no-same-owner --no-same-permissions $options") )
  {
    print "error: failed to open pipes to bsdtar\n";
  }
  # This section was contributed by Evangelos Foutras (foutrelis) to
  # solve the problem of pipes blocking due to buffer limitations.
  # The previous solution (read: kludge) involved multiple pipes to "cat".
  if (not defined(my $child_pid = fork()))
  {
    print "error: fork() failed\n";
  }
  elsif ($child_pid) # Parent
  {
    close $pin;
    my $output=<$pout>;
    waitpid $child_pid,0;
    waitpid $pid,0;
    return $output;
  }
  else # Child
  {
    close $pout;
    binmode $pin;
    print $pin $input;
    close $pin;
    exit(0);
  }
}

1;

__END__

=head1 NAME

Xyne::Arch::Rebase - A class to retrieve and update pacman package databases.

=head1 SYNOPSIS

  use Xyne::Arch::Rebase;

=head1 DESCRIPTION

This class is capable of retrieving archived pacman package databases from
mirror servers and extracting them to update the pacman sync database.
Extraction may be partial and limited to only those files which have changed.

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.