# Copyright (C) 2009  Xyne
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.


package Xyne::Arch::Reflector;
use strict;
use warnings;

use LWP::Simple qw/get/;
use POSIX qw/uname/;
use Xyne::Common::Functions qw/strip_trailing_slash/;

our $VERSION = '0.12';

my $MIRRORCHECK_URL = 'https://users.archlinux.de/~gerbra/mirrorcheck.html';
my $OPTIONS = {
  '--all'     => 'parametric',
  '--fastest' => 'parametric',
  '--help'    => 'switch',
  '--hours'   => 'parametric',
  '--latest'  => 'parametric',
  '--output'  => 'parametric',
  '--rank'    => 'switch'
};

my $EXPANSIONS = {
  '-a' => '--all',
  '-f' => '--fastest',
  '-h' => '--hours',
  '-l' => '--latest',
  '-o' => '--output',
  '-r' => '--rank'
};



sub new {
  my $proto = shift;
  my $class = ref($proto) || $proto;
  my $self  = {};
  bless ($self, $class);
  $self->{ARCH} = [uname()]->[4];
  return $self;
}

sub display_help
{
  print <<'HELP';
USAGE
  -a, --all
    return all servers (default)
    
  -f, --fastest n
    return fastest n servers as ranked by rankmirrors
    
  --help
    display this help message
    
  -h, --hours n
    return servers which have synced in the last n hours
    n may be a decimal number
    
  -l, --latest n
    return the n most recently synced servers
    
  -o, --output
    save mirrorlist to a file
    
  -r, --rank
    rank mirrors using rankmirrors

"--hours" accepts decimals

See the man page for more information.

HELP
}

sub get_repos
{
  my ($self) = @_;
  return keys %{ $self->{REPOS} };
}

sub get_mirrorcheck
{
  my ($self) = @_;
  my $mirrorcheck_html = get($MIRRORCHECK_URL);
  my $ignore = 1;
  my $unchecked = 0;
  my ($server,$repo,$arch);
  my $mirrorcheck = {};
  return $mirrorcheck if not defined($mirrorcheck_html);
  foreach my $line (split /\s*\n\s*/, $mirrorcheck_html)
  {
    $ignore = 0 if $line eq '<caption>All mirrors</caption>';
    next if $ignore;
    if ($line =~ m/href="([^"]+)"/)
    {
      my $url = $1;
      $server = $url if $url !~ m/^http:\/\/bugs.archlinux.org\//;
      $unchecked = 0;
    }
    elsif ($line =~ m/<td>([^-]*)-([^<]*)<\/td>/)
    {
      ($repo,$arch) = ($1,$2);
    }
    elsif ($line =~ m/UTC 1970<\/td>/)
    {
      $unchecked = 1;
    }
    elsif ($line =~ m/<td>(\d+) day\(s\), (\d+):(\d+) \(h:m\)/ and not $unchecked)
    {
      my $interval = $1*24 + $2 + $3/60;
      $mirrorcheck->{$server}->{$arch}->{$repo} = $interval;
      $self->{REPOS}->{$repo} ++;
      if (
        not defined($mirrorcheck->{$server}->{$arch}->{'AGE'}) or
        $mirrorcheck->{$server}->{$arch}->{'AGE'} < $interval
      )
      { $mirrorcheck->{$server}->{$arch}->{'AGE'} = $interval; }
    }
  }
  return $mirrorcheck;
}

sub get_mirrorlist
{
  my ($self,@servers) = @_;
  @servers = $self->get_servers if not @servers;
  my $mirrorlist = '';
  my $arch = $self->{ARCH};
  $mirrorlist .= "Server = $_\n" foreach @servers;
  return $mirrorlist;
}

sub get_servers
{
  my ($self) = @_;
  my $mirrorcheck = $self->get_mirrorcheck;
  my $arch = $self->{ARCH};
  my @servers = grep {defined($_)} keys %{ $mirrorcheck };
  if (defined($self->{ARGS}->{'--hours'}))
  {
    my $hours = $self->{ARGS}->{'--hours'};
    @servers = grep {$mirrorcheck->{$_}->{$arch}->{'AGE'}<=$hours} grep {defined($mirrorcheck->{$_}->{$arch}->{'AGE'})} @servers;
  }
  if (defined($self->{ARGS}->{'--latest'}))
  {
    my $latest = $self->{ARGS}->{'--latest'};
    @servers = sort {$mirrorcheck->{$a}->{$arch}->{'AGE'}<=>$mirrorcheck->{$b}->{$arch}->{'AGE'}} grep {defined($mirrorcheck->{$_}->{$arch}->{'AGE'})} @servers;
    @servers = grep {$_} @servers[0..($latest-1)];
  }
  @servers = map {&strip_trailing_slash($_)} @servers;

  # yeah, this is ugly
  if ($self->{ARGS}->{'--rank'} or $self->{ARGS}->{'--fastest'})
  {
    my $rmver = `rankmirrors --version 2>/dev/null`;
    if (length($rmver) > 0)
    {
      my $mirrorlist = join "\n", map {"Server = $_/\$repo/os/$arch"} @servers;
      my $n = $self->{ARGS}->{'--fastest'} ? '-n '.$self->{ARGS}->{'--fastest'} : '';
      open (my $rh,'-|',"rankmirrors $n - <<'MIRRORLIST'\n$mirrorlist\nMIRRORLIST\n");
      @servers = ();
      $|++;
      while (defined(my $line = <$rh>))
      {
        push @servers, $1 if $line =~ m/Server = (.*?)\/\$repo\/os\/$arch$/g;
      }
      close $rh;
    }
    else
    {
      print "# warning: failed to run rankmirrors... make sure that Python is installed\n";
    }
  }
  return map {"$_/\$repo/os/$arch"} @servers;
}

sub parse_args
{
  my ($self,@args) = @_;
  while (@args)
  {
    my $arg = shift @args;
    next if not defined($arg);
    $arg = $EXPANSIONS->{$arg} if $EXPANSIONS->{$arg};
    next if not $OPTIONS->{$arg};
    if ($OPTIONS->{$arg} eq 'parametric')
    {
      my $parameter = @args ? shift @args : '';
      $self->{ARGS}->{$arg} = $parameter;
    }
    else
    {
      $self->{ARGS}->{$arg}++;
    }
  }
  return $self;
}

sub run
{
  my ($self,@args) = @_;
  $self->parse_args(@args);
  if ($self->{ARGS}->{'--help'})
  {
    $self->display_help;
    exit;
  }
  my ($second,$minute,$hour,$day,$month,$year)=localtime(time);
  my $timestamp = sprintf("%4d-%02d-%02d %02d:%02d ",$year+1900,$month+1,$day,$hour,$minute);
  
  my $mirrorlist = "# $timestamp\n";
  $mirrorlist .= "# generated by reflector\n";
  $mirrorlist .= "# ranked by rankmirrors\n" if defined($self->{ARGS}->{'--rank'}) or defined($self->{ARGS}->{'--fastest'});
  
  
  $mirrorlist .= $self->get_mirrorlist();
  my $file = $self->{ARGS}->{'--output'};
  if (defined($file))
  {
    open (my $fh,'>',$file) or die "invalid output file specified: $!";
    print $fh $mirrorlist;
    close $fh;
  }
  else
  {
    print $mirrorlist;
  }
}

sub set_arch
{
  my ($self,$arch) = @_;
  $self->{ARCH} = $arch if defined($arch);
}

1;

__END__

=head1 NAME

Xyne::Arch::Reflector - A class to retrieve and filter the latest Arch Linux mirror list.

=head1 SYNOPSIS

  use Xyne::Arch::Reflector;

=head1 DESCRIPTION

A class that is capable of retrieving information from Gerhard Brauer's (GerBra)
Archlinux mirrorcheck page (http://users.archlinux.de/~gerbra/mirrorcheck.html).

=head2 EXPORT

None by default.

=head1 AUTHOR

Xyne <lt>xyne at archlinux dot ca<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2009  Xyne

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
